"""Runtimes module.

This module comprises classes for executing so called `RuntimeTasks` in `Runtimes` by leveraging the power of ssh.
The `RuntimeTask` class is a container for defining a sequence of elegantly task steps. This `RuntimeTask` can then be
executed either standalone of by passing it over to a `Runtime` instance. Passwordless ssh should be configured for all
hosts that should act as a `Runtime` to be able to conveniently manage those entities.

"""
import tempfile
import time

from fabric import Connection
from multiprocessing import Process, Manager
from typing import Optional, List, Dict, Generator, Union
import shutil
import cloudpickle as pickle
import atexit

import lazycluster.settings as settings

from lazycluster import InvalidRuntimeError, NoPortsLeftError, PortInUseError
import lazycluster._utils as _utils

import socket
import warnings

import json
import os
import psutil
import sys
import platform
import subprocess


class RuntimeTask(object):
    """This class provides the functionality for executing a sequence of elementary operations over ssh. The `fabric`
    library is used for handling ssh connections. A `RuntimeTask` can be composed from four different operations which
    we call steps, namely adding a step for running a shell command via `run_command()`, sending a file to a host via
    `send_file()`, retrieving a file from a host via `get_file()` or adding a step for executing a python function on a
    host via `run_function()`. The current restriction for running functions is that these functions need to be
    serializable via cloudpickle. To actually execute a `RuntimeTask`, i.e. the sequence of task steps, either a call
    to `execute()` is necessary or a handover to the `execute_task()` method of the `Runtime` class is necessary.
    Usually, a `RuntimeTask` will be executed in a `Runtime` or in a `RuntimeGroup`. See its documentation for further
    details.

    Examples:
        >>> # 1. Define a function that should be executed remotely via a RuntimeTask
        >>> def print():
        ...     print('Hello World!')

        >>> # 2. Create & compose the RuntimeTask by using the elementary operations
        >>> my_task = RuntimeTask('my-task').run_command('echo Hello World!').run_function(print)

        >>> # 3. Execute the RuntimeTask standalone w/o Runtime by handing over a fabric ssh connection
        >>> from fabric import Connection
        >>> task = my_task.execute(Connection('host'))

        >>> # 4. Check the logs of the RuntimeTask execution
        >>> task.print_log()
        >>> log = task.execution_log
    """

    def __init__(self, name: Optional[str] = None):
        """Initialization method.
        
        Args:
            name (Optional[str]): The name of the task. Defaults to None and consequently a unique identifier is
                                  generated via Python's id() function.
        """
        self._name = name
        if not self._name:
            self._name = str(id(self))
        self._task_steps = []
        self._execution_log = []
        self._function_return_pkl_paths = []  # file paths to files with function's pickled return data
        self._requested_files = []

        self._process = None

        # Will be created if run_function is executed      
        self._temp_dir = None
        # Cleanup will be done atexit since usage of destructor may lead to exceptions
        atexit.register(self.cleanup)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        return type(self).__name__ + ': ' + self.name

    def cleanup(self):
        """Remove temporary used resources, like temporary directories if created."""
        if self._temp_dir:
            shutil.rmtree(self._temp_dir)
            print('Temporary directory ' + self._temp_dir + ' of RuntimeTask ' + self.name + ' on localhost removed.')

    @property
    def name(self) -> str:
        """Get the task name.
        
        Returns:
            str: Task name
        """
        return self._name

    @property
    def execution_log(self) -> List[str]:
        """Return the execution log as list. The list is empty as long as a task was not yet executed. Each log entry
        corresponds to a single task step and the log index starts at 0. If th execution of an individual step does not
        produce and outut the list entry will be empty.
        """
        return self._execution_log

    @property
    def function_returns(self) -> Generator[object, None, None]:
        """Get the return data produced by functions which were executed as a consequence of a `task.run_function()`
        call.

       Internally, a function return is saved as a pickled file. This method unpickles each file one after
       another and yields the data. Moreover, the return data will be yielded in the same order as the functions were
       executed.

        Yields:
            Generator[object, None, None]: Generator object yielding the return data of the functions executed during
                                           task execution.
        """
        if self.process:
            self.process.join()

        for return_pkl_path in self._function_return_pkl_paths:

            if not os.path.exists(return_pkl_path):
                warnings.warn('Pickle file ' + return_pkl_path + ' with function return data does not exist. '
                              'Please check the logs.')
                yield None
                continue

            with open(return_pkl_path, 'rb') as file:
                yield pickle.load(file)

    @property
    def process(self) -> Process:
        """Returns the process object in which the task were executed. None, if not yet or synchronously executed. """
        return self._process

    def send_file(self, local_path: str, remote_path: Optional[str] = None):
        """Create a task step for sending either a single file or a folder from localhost to another host.
        
        Args:
            local_path (str): Path to file on local machine.
            remote_path (Optional[str]): Path on the remote host. Defaults to the root directory.
        
        Raises:
            ValueError: If file locally not found.
        """
        if not local_path:
            raise ValueError("Invalid local path")

        if not remote_path:
            remote_path = local_path

        self._task_steps.append(self._TaskStep(self._TaskStep.TYPE_SEND_FILE, local_path=local_path,
                                               remote_path=remote_path))
        return self

    def get_file(self, remote_path: str, local_path: Optional[str] = None):
        """Create a task step for getting either a single file or a folder from another host to localhost.

        Args:
            remote_path (str): Path to file on host.
            local_path (Optional[str]): Path to file on local machine. The remote file is downloaded 
                                        to the current working directory (as seen by os.getcwd) using 
                                        its remote filename if local_path is None. This is the default
                                        behavior of fabric.
        Raises:
            ValueError: If remote path is emtpy.
        """
        if not remote_path:
            raise ValueError("Remote path must not be empty")

        self._task_steps.append(self._TaskStep(self._TaskStep.TYPE_GET_FILE, remote_path=remote_path,
                                               local_path=local_path))
        self._requested_files.append(local_path)
        return self

    def run_command(self, command: str):
        """Create a task step for running a given shell command. 

        Args:
            command (str): Shell command.

        Raises:
            ValueError: If command is emtpy.
        """
        if not command:
            raise ValueError("Command must not be emtpy")
        self._task_steps.append(self._TaskStep(self._TaskStep.TYPE_RUN_COMMAND, command=command))

        return self

    def run_function(self, function: callable, **func_kwargs):
        """Create a task step for executing a given python function on a remote host. The function will be transferred
        to the remote host via ssh and cloudpickle. The return data can be requested via the property `function_returns`

        Note:
            Hence, the function must be serializable via cloudpickle and all dependencies must be available in its
            correct versions on the remote host for now. We are planning to improve the dependency handling.

        Args:
            function (callable): The function to be executed remotely.
            **func_kwargs: kwargs which will be passed to the function.
        
        Raises:
            ValueError: If function is empty.
        """
        if not function:
            raise ValueError("Function is invalid.")

        # Create a temp directory for persisting the serialized function
        if not self._temp_dir:
            self._temp_dir = tempfile.mkdtemp()
            self.run_command('mkdir -p ' + self._temp_dir)
            print('Temporary directory ' + self._temp_dir + ' for task ' + self.name + ' on localhost created.')

        # We distinguish between local -  and remote paths because in case we execute a function in a
        # `Runtime.create_instance('localhost')`, we may overwrite the files.
        input_function_name = function.__name__
        remote_file_name = self._create_file_name(input_function_name)
        local_file_name = 'local_' + remote_file_name
        local_pickle_file_path = os.path.join(self._temp_dir, local_file_name)
        remote_pickle_file_path = os.path.join(self._temp_dir, remote_file_name)

        local_return_file_name = 'return_' + remote_file_name
        remote_return_file_name = 'remote_' + local_return_file_name
        local_return_pickle_file_path = os.path.join(self._temp_dir, local_return_file_name)
        remote_return_pickle_file_path = os.path.join(self._temp_dir, remote_return_file_name)

        # Hereby we can pickle the function incl. kwargs, since we are
        # going to pickle the wrapper and not only the function itself.
        def function_wrapper():

            result = function(**func_kwargs)

            # Make the function return data available
            # => pickle the data and save it as a file on the remote host so that we can get the file back to the
            # local machine. Local machine means NOT the one where `function_wrapper()` is actually executed, but
            # the one where `task.run_function()` is actually executed.
            with open(remote_return_pickle_file_path, 'wb') as fp:
                pickle.dump(result, fp)
                fp.close()

        # Create cloud pickle file of the function_wrapper() and save it locally
        with open(local_pickle_file_path, 'wb') as f:
            pickle.dump(function_wrapper, f)
            f.close()

        # Add `TaskStep` for sending the pickled function_wrapper file to the Runtime
        self.send_file(local_pickle_file_path, remote_pickle_file_path)

        # Ensure installation of cloudpickle 
        self.run_command('pip install -q cloudpickle')

        # Create commands for executing the pickled function file
        run_function_code = 'import cloudpickle as pickle; file=open(\'' + remote_pickle_file_path + '\', \'rb\'); ' \
                            'fn=pickle.load(file); file.close(); fn()'
        run_cmd = 'python -c "' + run_function_code + '"'
        self.run_command(run_cmd)

        # Ensure the cleanup of the created function pickle file on the Runtime
        run_cmd = 'rm ' + remote_pickle_file_path
        self.run_command(run_cmd)

        # Get the return data back as pickled file
        self.get_file(remote_return_pickle_file_path, local_return_pickle_file_path)
        self._function_return_pkl_paths.append(local_return_pickle_file_path)

        # Cleanup the return pickle file on the remote host
        run_cmd = 'rm ' + remote_return_pickle_file_path
        self.run_command(run_cmd)

        return self

    def execute(self, connection: Connection):
        """Execute the task on a remote host using a fabric connection.
        
        Args:
            connection (fabric.Connection): Fabric connection object managing the ssh connection to the remote host.
        Raises:
            ValueError: If cxn is broken and connection can not be established.
        """
        # Fabric will only consider the manually set root directory for connection.run()
        # but not fur connection.put() & connection.get(). Thus, we need to set the path in
        # these cases manually.
        try:
            root_dir = connection.run('pwd').stdout.replace('\n', '').replace('\r', '')
        except socket.gaierror:
            raise ValueError('Connection cannot be established. connection: ' + str(connection))

        for task_step in self._task_steps:

            if task_step.type == self._TaskStep.TYPE_RUN_COMMAND:
                res = connection.run(task_step.command, pty=True)
                stdout = res.stdout.replace('\n', '').replace('\r', '')
                self._execution_log.append(stdout)

            elif task_step.type == self._TaskStep.TYPE_SEND_FILE:
                remote_path = task_step.remote_path
                if not remote_path:
                    # Manually set the root dir since fabric defaults to the os user's home dir
                    local_dir, filename = os.path.split(task_step.local_path)
                    remote_path = os.path.join(root_dir, filename)
                connection.put(task_step.local_path, remote_path)
                self._execution_log.append('Send file ' + task_step.local_path + ' to ' + remote_path + connection.host)

            elif task_step.type == self._TaskStep.TYPE_GET_FILE:
                # Set root dir manually, so that it feels naturally for the user if he just 
                # uses the filename w/o a path
                remote_path = task_step.remote_path
                if not remote_path.startswith('/'):
                    remote_dir, filename = os.path.split(remote_path)
                    remote_path = os.path.join(root_dir, filename)
                connection.get(remote_path, task_step.local_path)
                self._execution_log.append('Get file ' + task_step.local_path + ' to remote.')

    def join(self):
        """Block the execution until the `RuntimeTask` finished its asynchronous execution. """
        if self.process:
            self.process.join()

    def print_log(self):
        """Print the execution log. Each log entry will be printed separately. The log index will be prepended."""
        if not self.execution_log:
            print('The log of task ' + self.name + 'is empty!')
        else:
            print('Log of Task ' + self.name + ':')
        i = 0
        for entry in self.execution_log:
            print(str(i) + ': ' + entry)
            i += 1

    # Private helper methods & attributes
    _function_index = 0
    _FUNCTION_PREFIX = 'func'
    _PICKLE_FILENAME_EXT = 'pkl'

    @classmethod
    def _create_file_name(cls, function_name: Optional[str] = None):
        """Creates a program unique function name. """
        if function_name:
            file_prefix = function_name
        else:
            file_prefix = cls._FUNCTION_PREFIX
        cls._function_index += 1
        return file_prefix + str(cls._function_index) + '.' + cls._PICKLE_FILENAME_EXT

    class _TaskStep:
        """Represents an individual action, i.e. a `_TaskStep` within a `RuntimeTask`. """
        TYPE_RUN_COMMAND = 'command'
        TYPE_SEND_FILE = 'send-file'
        TYPE_GET_FILE = 'get-file'

        def __init__(self, step_type: str, local_path: Optional[str] = None,
                     remote_path: Optional[str] = None, command: Optional[str] = None):
            self.type = step_type
            self.local_path = local_path
            self.remote_path = remote_path
            self.command = command

        def __repr__(self):
            return "%s(%r)" % (self.__class__, self.__dict__)


class Runtime(object):
    """Runtime for executing `RuntimeTasks` in it or exposing ports from / to localhost.
    
    Note: Passwordless ssh access should be be setup in advance. Otherwise the connection kwargs of fabric must be used
          for setting up the ssh connection.

    Examples:
        Execute a RuntimeTask synchronously
        >>> Runtime('host-1').execute_task(my_task, execute_async=False)

        Expose a port from localhost to the remote host so that a service running on localhost
        is accessible from the remote host as well.
        >>> Runtime('host-1').expose_port_to_runtime(8786)

        Expose a port from a remote `Runtime` to to localhost so that a service running on the `Runtime`
        is accessible from localhost as well.
        >>> Runtime('host-1').expose_port_from_runtime(8787)
    """

    LOCALHOST = 'localhost'

    # Constants used for generating process keys
    _PORT_TO_RUNTIME = '-R'
    _PORT_FROM_RUNTIME = '-L'
    _PROCESS_KEY_DELIMITER = '::'
    _TASK_PROCESS_KEY_PREFIX = 'task'

    def __init__(self, host: str, root_dir: Optional[str] = None, **connection_kwargs):
        """Initialization method.

        Args:
            host (str): The host of the `Runtime`.
            root_dir (Optional[str]): The directory which shall act as root one. Defaults to None.
                                      Consequently, a temporary directory will be created and used as root directory. If
                                      the root directory is a temporary one it will be cleaned up either `atexit` or
                                      when calling `cleanup()` manually.

            **connection_kwargs: kwargs that will be passed on to the fabric connection. Please check the fabric docs
                                 for further details.

        Raises:
            InvalidRuntimeError: If `Runtime` could not be instantiated successfully.
        """

        self._host = host
        self._connection_kwargs = connection_kwargs
        self._root_dir = root_dir
        self._root_dir_is_temp = False  # Indicates that a temp directory acts as root, which will be removed at the end
        self._processes = {}  # The dict key is a generated process identifier and the value contains the process
        self._info = {}
        self._process_manager = Manager()
        self._tasks = []

        # Cleanup will be done atexit since usage of destructor may lead to exceptions
        atexit.register(self.cleanup)

        # The check shall be executed at the end of the method, since it relies on some attributes like _root_dir etc.
        if not self.is_valid_runtime():
            raise InvalidRuntimeError(host)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        return self.class_name + ': ' + self.host

    @property
    def root_directory(self):
        """The path of the root directory that was set during object initialization. """
        self._create_root_dir_if_not_exists()
        return self._root_dir

    @property
    def task_processes(self) -> List[Process]:
        """Get all processes that were started to execute a `RuntimeTask` asynchronously.

        Returns:
            List[Process]: RuntimeTask processes.
        """
        processes = []
        for process_key, process in self._processes.items():
            if self.is_runtime_task_process(process_key):
                processes.append(process)
        return processes

    @property
    def function_returns(self) -> Generator[object, None, None]:
        """Get the return data produced by Python functions which were executed as a consequence of
           `task.run_function()`. The call will be passed on to the `function_returns` property of the `RuntimeTask`.
           The order is determined by the order in which the `RuntimeTasks` were executed in the `Runtime`.

        Yields:
            Generator[object, None, None]: Generator object yielding the return data of the functions executed during
                                           task execution.
        """
        for task in self._tasks:
            for function_return in task.function_returns:
                yield function_return

    @classmethod
    def is_runtime_task_process(cls, process_key: str) -> bool:
        """Checks if the process which belongs to a given `process_key` was started to exectue a `RuntimeTask` based on
        an internal naming scheme of the process keys.

        Args:
            process_key (str): The generated process identifier.
        Returns:
            bool: True, if process was started to execute a `RuntimeTask`
        """
        key_splits = process_key.split(cls._PROCESS_KEY_DELIMITER)
        return True if key_splits[0] == cls._TASK_PROCESS_KEY_PREFIX else False

    @classmethod
    def is_port_exposure_process(cls, process_key: str) -> bool:
        """Check if the process which belongs to the given `process_key` is used for exposing a port, i.e. keeping
        an ssh tunnel alive.

        Args:
            process_key (str): The generated process identifier.
        Returns:
            bool: True, if process is used for port exposure.
        """
        key_splits = process_key.split(cls._PROCESS_KEY_DELIMITER)
        return True if key_splits[0] == cls._PORT_FROM_RUNTIME or key_splits[0] == cls._PORT_TO_RUNTIME else False

    @property
    def host(self) -> str:
        """Get the host of the runtime.

        Returns:
            str:  The host of the runtime.
        """
        return self._host

    @property
    def info(self) -> dict:
        """Get information about the runtime.

        Returns:
            dict: Runtime information.
        """
        if not self._info:
            self._info = self._read_info()
        return self._info

    @property
    def os(self) -> str:
        """Get operating system information.

        Returns:
            str: OS.  
        """
        if not self._info:
            self._info = self._read_info()
        return self._info['os']

    @property
    def cpu_cores(self) -> int:
        """Get information about the available CPUs. If you are in a container 
        the CPU quota will be given if set. Otherwise, the number of physical CPUs
        on the host machine is given.

        Returns:
            str: CPU quota.  
        """
        if not self._info:
            self._info = self._read_info()
        return int(self._info['cpu_cores'])

    @property
    def memory(self) -> int:
        """Get information about the total memory in bytes.

        Returns:
            str: Total memory in bytes.  
        """
        if not self._info:
            self._info = self._read_info()
        return int(self._info['memory'])

    @property
    def memory_in_mb(self):
        """Get the memory information in mb. """
        return self.memory / 1024 / 1024

    @property
    def python_version(self) -> str:
        """Get the python version.

        Returns:
            str: Python version.  
        """
        if not self._info:
            self._info = self._read_info()
        return self._info['python_version']

    @property
    def gpus(self) -> list:
        """GPU information as list. Each list entry contains information for one GPU.

        Returns:
            list: List with GPU information.
        """
        if not self._info:
            self._info = self._read_info()
        return self._info['gpus']

    @property
    def gpu_count(self) -> int:
        """The count of GPUs."""
        return len(self.gpus)

    @property
    def class_name(self) -> str:
        """Get the class name  as string. """
        return self.__class__.__name__

    @property
    def alive_process_count(self):
        """Get the number of alive processes. """
        return len(self.get_processes(only_alive=True))

    @property
    def alive_task_process_count(self):
        """Get the number of alive processes which were started to execute a `RuntimeTask`. """
        return len(self.task_processes)

    def is_valid_runtime(self) -> bool:
        """Checks if a given host is a valid `Runtime`.

        Returns:
            bool: True, if it is a valid remote runtime.
        """

        task = RuntimeTask('check-python-version')
        task.run_command('python --version')

        try:
            self.execute_task(task, execute_async=False)
        except Exception:
            return False

        if not task.execution_log[0]:
            return False

        # Example: `Python 3.6.8 :: Anaconda, Inc.`
        python_version = task.execution_log[0].split()[1].split('.')

        if not python_version:
            return False
        elif int(python_version[0]) > 3:  # Stay future-proof
            warnings.warn('The lib was originally created for Python 3.6 and is not yet fully tested for '
                          'Python >= Python 4')
            return True
        elif int(python_version[0]) == 3 and int(python_version[1]) >= 6:
            return True
        else:
            return False

    def get_free_port(self, ports: List[int]) -> int:
        """Returns the first port from the list which is currently not in use in the `Runtime`.

        Args:
             ports List[int]: The list of ports that will be used to check if the port is currently in use.

        Returns:
            int: The first port from the list which is not yet used within the whole group.

        Raises:
            NoPortsLeftError: If the port list is empty and no free port was found yet.
        """
        for port in ports:
            if self.has_free_port(port):
                return port

        raise NoPortsLeftError()

    def execute_task(self, task: RuntimeTask, execute_async: Optional[bool] = True):
        """Execute a given `RuntimeTask` in the `Runtime`.

        Args:
            task (RuntimeTask): The task to be executed.
            execute_async (bool, Optional): The execution will be done in a separate thread if True. Defaults to True.
        """
        self._create_root_dir_if_not_exists()
        async_str = ' asynchronously ' if execute_async else ' synchronously '
        print('Executing task ' + task.name + async_str + ' on ' + self._host)

        # Wrapper needed to ensure execution from root directory
        def execute_remote_wrapper():
            cxn = self._fabric_connection
            with cxn.cd(self._root_dir):
                task.execute(cxn)

        if execute_async:
            # Initialize logs with managed list (multiprocessing)
            # => We can access the logs although it is executed in another process
            task._execution_log = self._process_manager.list()
            process = Process(target=execute_remote_wrapper)
            process.start()
            self._processes.update({self._create_process_key_for_task_execution(task): process})
            task._process = process
        else:
            execute_remote_wrapper()

        self._tasks.append(task)

    def _create_root_dir_if_not_exists(self):
        if not self._root_dir:
            self._root_dir = self.create_tempdir()
            self._root_dir_is_temp = True
            print('Temporary directory ' + self._root_dir + ' created on ' + self._host)

    def print_log(self):
        """Print the execution logs of each `RuntimeTask` that was executed in the `Runtime`. """
        for task in self._tasks:
            task.print_log()

    def clear_tasks(self):
        """Clears all internal state related to `RuntimeTasks`. """
        self._tasks = []
        self._processes = {key: value for key, value in self._processes.items()
                           if not Runtime.is_runtime_task_process(key)}

    def expose_port_to_runtime(self, local_port: int, runtime_port: Optional[int] = None) -> str:
        """Expose a port from localhost to the `Runtime` so that all traffic on the `runtime_port` is forwarded to the
        `local_port` on localhost.
        
        Args:
            local_port (int): The port on the local machine.
            runtime_port (Optional[int]): The port on the runtime. Defaults to `local_port`.

        Returns:
            str: Process key, which can be used for manually stopping the process running the port exposure for example.
        
        Examples:
            A DB is running on localhost on port `local_port` and the DB is only accessible from localhost.
            But you also want to access the service on the remote `Runtime` on port `runtime_port`. Then you can use
            this method to expose the service which is running on localhost to the remote host.
        """
        if not runtime_port:
            runtime_port = local_port
        if local_port == runtime_port and self._host == self.LOCALHOST:
            # We can not forward a port to itself!
            print('Port exposure skipped. Can\'t expose a port to myself.')
            return ''
        elif not self.has_free_port(runtime_port):
            raise PortInUseError(runtime_port, runtime=self)

        proc = Process(target=self._forward_runtime_port_to_local,
                       kwargs={'local_port': local_port, 'runtime_port': runtime_port})
        proc.start()
        key = self._create_process_key_for_port_exposure(self._PORT_TO_RUNTIME, local_port, runtime_port)
        self._processes.update({key: proc})
        time.sleep(0.1)  # Necessary to prevent collisions with MaxStartup restrictions
        print('Local port ' + str(local_port) + ' exposed to runtime ' + self._host + ' on port ' +
              str(runtime_port))
        return key

    def expose_port_from_runtime(self, runtime_port: int, local_port: Optional[int] = None) -> str:
        """Expose a port from a `Runtime` to localhost so that all traffic to the `local_port` is forwarded to the
        `runtime_port` of the `Runtime`. This corresponds to local port forwarding in ssh tunneling terms.
        
        Args:            
            runtime_port (int): The port on the runtime.
            local_port (Optional[int]): The port on the local machine. Defaults to `runtime_port`.

        Returns:
            str: Process key, which can be used for manually stopping the process running the port exposure.
        
        Examples:
            A DB is running on a remote host on port `runtime_port` and the DB is only accessible from the remote host.
            But you also want to access the service from the local machine on port `local_port`. Then you can use this
            method to expose the service which is running on the remote host to localhost.

        """
        if not local_port:
            local_port = runtime_port
        if local_port == runtime_port and self._host == self.LOCALHOST:
            # We can not forward a port to itself!
            print('Port exposure skipped. Can\'t expose a port to myself.')
            return ''
        elif not localhost_has_free_port(local_port):
            raise PortInUseError(runtime_port, runtime=self)

        proc = Process(target=self._forward_local_port_to_runtime,
                       kwargs={'local_port': local_port, 'runtime_port': runtime_port})
        proc.start()
        key = self._create_process_key_for_port_exposure(self._PORT_FROM_RUNTIME, local_port, runtime_port)
        self._processes.update({key: proc})
        time.sleep(0.1)  # Necessary to prevent collisions with MaxStartup restrictions
        print('Port ' + str(runtime_port) + ' from runtime ' + self._host + ' exposed to local port ' +
              str(local_port))
        return key

    def get_process(self, key: str) -> Process:
        """Get an individual process by process key.
        
        Args:
            key (str): The key identifying the process.
            
        Returns:
            Process: The desired process.

        Raises:
            ValueError: Unknown process key.
        """
        if key not in self._processes:
            raise ValueError('Unknown process key.')
        return self._processes[key]

    def get_processes(self, only_alive: bool = False) -> Dict[str, Process]:
        """Get all managed processes or only the alive ones as dictionary with the process key as dict key. An
        individual process can be retrieved by key via `get_process()`.
        
        Args:
            only_alive (bool): True, if only alive processes shall be returned instead of all. Defaults to False.
            
        Returns:
            Dict[str, Process]: Dictionary with process keys as dict keys and the respective processes as dict values.
        """
        if only_alive:

            ret_processes = {}

            for key, process in self._processes.items():
                if process.is_alive():
                    ret_processes.update({key: process})
            return ret_processes

        else:
            return self._processes

    def stop_process(self, key: str):
        """Stop a process by its key. 

        Args: 
            key (str): The key identifying the process.
        
        Raises:
            ValueError: Unknown process key.
        """
        if key not in self._processes:
            raise ValueError('Unknown process key.')
        self._processes[key].terminate()

    def has_free_port(self, port: int) -> bool:
        """Checks if the port is available on the runtime. 
        
        Args:
            port (int): The port which will be checked.

        Returns:
            bool: True if port is free, else False.
        """
        with self._fabric_connection as cxn:
            cmd_str = 'python -c "import socket;print(\'free\') ' \
                      'if socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect_ex((\'localhost\', ' + str(port) + \
                      ')) else None"'
            res = cxn.run(cmd_str)
            return True if res.stdout else False

    def print_info(self):
        """Print the Runtime info formatted as table."""
        info = self.info
        print('\n')
        print('Information of `' + self.class_name + '` ' + self.host + ':')
        for key, value in info.items():
            display_value = value if not key == 'memory' else str(self.memory_in_mb) + ' mb'
            print('{:<8} {:<8}'.format(key, display_value))

    def check_filter(self,
                     gpu_required: bool = False,
                     min_memory: Optional[int] = None,
                     min_cpu_cores: Optional[int] = None,
                     installed_executables: Union[str, List[str], None] = None,
                     filter_commands: Union[str, List[str], None] = None) -> bool:
        """Checks the `Runtime` object for certain filter criteria.

        Args:
            gpu_required (bool): True, if gpu availability is required. Defaults to False.
            min_memory (Optional[int]): The minimal amount of memory in MB. Defaults to None, i.e. not restricted.
            min_cpu_cores (Optional[int]): The minimum number of cpu cores required. Defaults to None, i.e. not
                                           restricted.
            installed_executables (Union[str, List[str], None]): Possibility to check if an executable is
                                                                installed. E.g. if the executable `ping` is installed.
            filter_commands (Union[str, List[str], None]): Shell commands that can be used for generic filtering.
                                                           See examples. A filter command must echo true to be evaluated
                                                           to True, everything else will be interpreted as False.
                                                           Defaults to None.

        Returns:
            bool: True, if all filters were successfully checked otherwise False.

        Examples:
            >>> runtime = Runtime('host-1')
            Check if the `Runtime` has a specific executable installed
            such as `ping` the network administration software utility.
            >>> check_passed = runtime.check_filter(installed_executables='ping')

            Check if a variable `WORKSPACE_VERSION` is set on the `Runtime`
            >>> filter_str = '[ ! -z "$WORKSPACE_VERSION" ] && echo "true" || echo "false"'
            >>> check_passed = runtime.check_filter(filer_commands=filter_str)
        """

        all_filters_checked = True

        if gpu_required and not self.gpus:
            print(self.class_name + ' ' + self.host + 'does not have GPUs.')
            all_filters_checked = False

        if min_memory and self.memory < min_memory:
            print(self.class_name + ' ' + self.host + 'has only ' + str(self.memory) + ' instead of ' + str(min_memory) +
                  ' as required')
            all_filters_checked = False

        if min_cpu_cores and self.cpu_cores < min_cpu_cores:
            print(self.class_name + ' ' + self.host + 'has only ' + str(self.cpu_cores) + ' instead of ' + str(min_cpu_cores) +
                  ' as required.')
            all_filters_checked = False

        if installed_executables:
            for executable_name in _utils.create_list_from_parameter_value(installed_executables):
                if not self._has_executable_installed(str(executable_name)):
                    print(self.class_name + ' ' + self.host + 'does not have executable ' + str(executable_name) +
                          'installed.')
                    all_filters_checked = False

        if filter_commands:
            for filter_command in _utils.create_list_from_parameter_value(filter_commands):
                if not self._filter_command_checked(str(filter_command)):
                    print('Filter ' + filter_commands + ' could not be checked successfully on ' + self.class_name + ' ' +
                          self.host + ' .')
                    all_filters_checked = False

        return all_filters_checked

    def create_tempdir(self) -> str:
        """Create a temporary directory and return its name/path.

        Returns:
            str: The name/path of the directory.
        """
        with self._fabric_connection as cxn:
            cmd_str = 'python -c "import tempfile; print(tempfile.mkdtemp())"'
            res = cxn.run(cmd_str)
            path = res.stdout.split("\n")[0]
            if not path:
                path = res.stdout
            return path

    def create_dir(self, path):
        """Create a directory. All folders in the path will be created if not existing.

        Args:
            path (str): The full path of the directory to be created.
        """
        with self._fabric_connection as cxn:
            cmd_str = 'mkdir -p ' + path
            res = cxn.run(cmd_str)

    def delete_dir(self, path: str) -> bool:
        """Delete a directory recursively. If at least one contained file could not be removed then False is returned.

        Args:
            path (str): The full path of the directory to be deleted.

        Returns:
            bool: True if the directory could be deleted successfully.
        """
        with self._fabric_connection as cxn:
            cmd_str = 'rm -r ' + path
            try:
                cxn.run(cmd_str)
                return True
            except:
                return False

    def cleanup(self):
        """Release all acquired resources and terminate all processes. """
        for key, process in self._processes.items():
            process.terminate()
            process.join()
            if process.is_alive():
                print('Process with key ' + key + ' could not be terminated')
            else:
                print('Process with key ' + key + ' terminated')
        if self._root_dir_is_temp and self._root_dir:
            success = self.delete_dir(self._root_dir)
            if success:
                print('Temporary directory ' + self.root_directory + ' of Runtime ' + self.host + ' removed.')
                self._root_dir = None
                self._root_dir_is_temp = False
            else:
                print('Temporary directory ' + self.root_directory + ' of Runtime ' + self.host +
                      ' could not be removed.')

    # - Private methods -#

    @classmethod
    def _create_process_key_for_port_exposure(cls, direction: str, local_port: int, runtime_port: int) -> str:
        """Create a process key for processes exposing ports, i.e. keeping ssh tunnels open.
        This key will act as an identifier for internally generated processes.
        
        Raises:
            ValueError: If direction has an invalid value.
        """
        if not local_port:
            local_port = runtime_port
        if not runtime_port:
            runtime_port = local_port

        delimiter = cls._PROCESS_KEY_DELIMITER

        if direction == cls._PORT_FROM_RUNTIME:
            return cls._PORT_FROM_RUNTIME + delimiter + str(local_port) + delimiter + str(runtime_port)
        elif direction == cls._PORT_TO_RUNTIME:
            return cls._PORT_TO_RUNTIME + delimiter + str(runtime_port) + delimiter + str(local_port)
        else:
            raise ValueError(direction + ' is not a supported runtime process prefix type')

    @classmethod
    def _create_process_key_for_task_execution(cls, task: RuntimeTask) -> str:
        """Create a process key for processes started to execute a `RuntimeTasks` asynchronously
        This key will act as an identifier for internally generated processes.
        """
        return cls._TASK_PROCESS_KEY_PREFIX + cls._PROCESS_KEY_DELIMITER + task.name

    @classmethod
    def _create_executable_installed_shell_cmd(cls, executable: str) -> str:
        return 'hash ' + executable + ' 2>/dev/null && echo "true" || echo ""'

    def _has_executable_installed(self, executable_name: str) -> bool:
        """Checks if an executable is installed on the runtime."""
        shell_cmd = self._create_executable_installed_shell_cmd(executable_name)
        return self._filter_command_checked(shell_cmd)

    def _filter_command_checked(self, shell_cmd: str) -> bool:
        task = RuntimeTask('_filter_command_checked')
        task.run_command(shell_cmd)
        self.execute_task(task, execute_async=False)

        # Check the last log entry for the string true
        result = str(task.execution_log[len(task.execution_log) - 1])
        return True if result.lower() == 'true' else False

    @property
    def _fabric_connection(self) -> Connection:
        """Get a new fabric connection to the runtime.

        Raises:
            ValueError: If user or port values are given via both host shorthand
                        and their own arguments.
        """
        try:
            return Connection(host=self.host, connect_kwargs=self._connection_kwargs)
        except socket.gaierror:
            raise ValueError('Cannot establish SSH connection to host ' + self.host)

    def _forward_local_port_to_runtime(self, local_port: int, runtime_port: Optional[int] = None):
        """Creates ssh connection to the runtime and creates then a ssh tunnel
        from `localhost`:`local_port to `runtime`:`runtime_port`. """
        if not runtime_port:
            runtime_port = local_port

        with self._fabric_connection.forward_local(local_port, runtime_port):
            while True:
                time.sleep(1000)

    def _forward_runtime_port_to_local(self, runtime_port: int, local_port: Optional[int] = None):
        """Creates ssh connection to the runtime and then creates a ssh tunnel
        from `runtime`:`runtime_port` to `localhost`:`local_port`. """
        if not local_port:
            local_port = runtime_port

        with self._fabric_connection.forward_remote(runtime_port, local_port):
            while True:
                time.sleep(1000)

    def _read_info(self) -> dict:
        """Read the host machine information.  """
        task = RuntimeTask('get-host-info')
        task.run_command('pip install -q --upgrade ' + settings.PIP_PROJECT_NAME)
        task.run_function(print_localhost_info)
        self.execute_task(task, execute_async=False)
        return json.loads(task.execution_log[4])


""" Module level utils """


def get_localhost_info() -> dict:
    """Get information about the specifications of localhost.

    Returns:
        dict: Current dict keys: 'os', 'cpu_cores', 'memory', 'python_version', 'workspace_version', 'gpus'.
    """
    info = {
        'os': _get_os_on_localhost(),
        'cpu_cores': _get_cpu_count_on_localhost(),
        'memory': _get_memory_on_localhost(),
        'python_version': _get_python_version_on_localhost(),
        'workspace_version': _get_workspace_version_on_localhost(),
        'gpus': _get_gpu_count_from_localhost()
    }
    return info


def print_localhost_info():
    """Prints the dictionary retrieved by `get_localhost_info()`."""
    json_str = json.dumps(get_localhost_info())
    print(json_str)


def command_exists_on_localhost(command: str) -> bool:
    """Check if a command exists on localhost"""
    cmd_ext = 'hash ' + command + ' 2>/dev/null && echo "True" || echo ""'
    return True if os.popen(cmd_ext).read() else False


def localhost_has_free_port(port: int) -> bool:
    """Checks if the port is free on localhost.

    Args:
        port (int): The port which will be checked.
    Returns:
        bool: True if port is free, else False.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    is_success = sock.connect_ex(('localhost', port))
    return True if is_success else False


""" Private module level utils """


def _get_os_on_localhost() -> str:
    return platform.platform()


def _get_cpu_count_on_localhost() -> str:
    # Read the cpu quota
    cpu_quota = os.popen('cat /sys/fs/cgroup/cpu/cpu.cfs_quota_us').read()
    # Remove newlines
    cpu_quota = cpu_quota.replace('\n', '')
    if not cpu_quota.isnumeric():
        return str(os.cpu_count())
    else:
        return str(float(cpu_quota) / 100000)


def _get_memory_on_localhost() -> str:
    mem_limit = None
    if os.path.isfile("/sys/fs/cgroup/memory/memory.limit_in_bytes"):
        with open('/sys/fs/cgroup/memory/memory.limit_in_bytes', 'r') as file:
            mem_limit = file.read().replace('\n', '').strip()

    total_memory = psutil.virtual_memory().total

    if not mem_limit:
        mem_limit = total_memory
    elif int(mem_limit) > int(total_memory):
        # if mem limit from cgroup bigger than total memory -> use total memory
        mem_limit = total_memory

    return str(mem_limit)


def _get_python_version_on_localhost() -> str:
    return str(sys.version_info.major) + '.' + str(sys.version_info.minor) + '.' + str(sys.version_info.micro)


def _get_workspace_version_on_localhost() -> str:
    return os.environ['WORKSPACE_VERSION']


def _get_gpu_count_from_localhost() -> int:
    NVIDIA_CMD = 'nvidia-smi'

    if not command_exists_on_localhost(NVIDIA_CMD):
        return 0

    try:
        sp = subprocess.Popen([NVIDIA_CMD, '-q'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out_str = sp.communicate()
        out_list = out_str[0].decode("utf-8").split('\n')

        count_gpu = 0
        for item in out_list:
            try:
                key, val = item.split(':')
                key, val = key.strip(), val.strip()
                if key == 'Product Name':
                    count_gpu += 1
            except:
                pass
    except:
        count_gpu = 0
    return count_gpu



"""Module for conveniently managing a DASK cluster. http://distributed.dask.org """

from distributed import Client
import time
from typing import List, Union, Optional

from subprocess import Popen

from lazycluster import RuntimeTask, Runtime, RuntimeGroup, localhost_has_free_port
from lazycluster.cluster import MasterWorkerCluster, MasterLauncher, WorkerLauncher
from lazycluster import _utils
from lazycluster.cluster.exceptions import MasterStartError
from lazycluster.exceptions import PortInUseError


class LocalMasterLauncher(MasterLauncher):
    """Concrete implementation of the `MasterLauncher` interface.

    This class implements the logic for starting a the DASK master instance (i.e. scheduler in DASK terms) on localhost.
    """

    def start(self, ports: Union[List[int], int], timeout: int = 3) -> List[int]:
        """Launch a master instance.

        Args:
            ports (Union[List[int], int]): Port where the master should be started. If a list is given then the
                                                 first port that is free in the `RuntimeGroup` will be used. The actual
                                                 chosen port can requested via the property `port`.
            timeout (int): Timeout (s) after which an MasterStartError is raised if master instance not started yet.

        Returns:
            List[int]: In case a port list was given the updated port list will be returned. Otherwise an empty list.

        Raises:
            PortInUseError: If a single port is given and it is not free in the `RuntimeGroup`.
            NoPortsLeftError: If a port list was given and none of the ports is actually free in the `RuntimeGroup`.
            MasterStartError: If master was not started after the specified `timeout`.
        """

        if not isinstance(ports, list):
            if localhost_has_free_port(ports) and self._group.has_free_port(ports, exclude_hosts=Runtime.LOCALHOST):
                master_port = ports
            else:
                raise PortInUseError(ports, self._group)

        else:
            master_port = self._group.get_free_port(ports)  # Raises NoPortsLeftError
            ports = _utils.get_remaining_ports(ports, master_port)

        self._process = Popen(['dask-scheduler', '--port', str(master_port)])

        time.sleep(timeout)  # Needed for being able to check the port

        if not localhost_has_free_port(master_port):
            self._port = master_port
            print('Dask scheduler started on localhost on port ' + str(self._port))
        else:
            raise MasterStartError('localhost', self._port)

        # Sets up ssh tunnel for scheduler such that all communication is routed over the
        # local machine and all entities can talk to each the scheduler on localhost.
        self._group.expose_port_to_runtimes(self._port)

        return ports


class RoundRobinLauncher(WorkerLauncher):
    """WorkerLauncher implementation for launching DASK workers in a round robin manner. """

    def __init__(self, runtime_group: RuntimeGroup):
        """Initialization method.

        Args:
            runtime_group (RuntimeGroup): The group where the workers will be started.
        """
        super().__init__(runtime_group)
        self._ports = None

    def start(self, worker_count: int, master_port: int, ports: List[int]) -> List[int]:
        """Launches the worker instances in the `RuntimeGroup`.

        Args:
            worker_count (int): The number of worker instances to be started in the group.
            master_port (int):  The port of the master instance.
            ports (List[int]): The ports to be used for starting the workers. Only ports from the list will be chosen
                               that are actually free.
        Returns:
            List[int]: The updated port list after starting the workers, i.e. the used ones were removed.

        Raises:
            NoPortsLeftError: If there are not enough free ports for starting all workers.
        """
        self._ports = ports
        hosts = self._group.hosts
        runtimes = self._group.runtimes

        # Launch each desired worker one by one
        for worker_index in range(worker_count):

            # Determine the runtime where the next worker will be started in
            runtime_index = (self._group.runtime_count + worker_index) % self._group.runtime_count
            # Get the actual host corresponding to the index
            host = hosts[runtime_index]
            root_dir = runtimes[runtime_index].root_directory
            assert host == runtimes[runtime_index].host

            worker_port = self._launch_single_worker(host, worker_index, master_port, root_dir)  # NoPortsLeftError
            # Remember which worker ports are now used per `Runtime`
            if host in self._ports_per_host:
                self._ports_per_host[host].append(worker_port)
            else:
                self._ports_per_host.update({host: [worker_port]})

        # Needed for worker communication between workers, which is necessary in case of DASK
        self.setup_worker_ssh_tunnels()

        return self._ports

    def _launch_single_worker(self, host: str, worker_index: int, master_port: int, root_directory: str):
        """Launch a single worker instance in a `Runtime` in the `RuntimeGroup`.

        Raises:
            NoPortsLeftError
        """
        # 1. Get a free port based on the port list
        worker_port = self._group.get_free_port(self._ports)  # Raises NoPortsLeftError
        self._ports = _utils.get_remaining_ports(self._ports, worker_port)
        # 2. Start the worker on this port
        task = RuntimeTask('launch-dask-worker-' + str(worker_index))
        task.run_command(DaskCluster.PIP_INSTALL_COMMAND)
        task.run_command(self._get_launch_command(master_port, worker_port, root_directory))
        self._group.execute_task(task, host)
        return worker_port

    @classmethod
    def _get_launch_command(cls, master_port: int, worker_port: int, root_directory: str) -> str:
        """Get the shell command for starting a worker instance.

        Returns:
             str: The launch command.
        """
        return 'dask-worker --worker-port=' + str(worker_port) + ' --local-directory=' + root_directory \
               + ' localhost:' + str(master_port)


class DaskCluster(MasterWorkerCluster):
    """Convenient class for launching a Dask cluster in a `RuntimeGroup`. 

    Examples:
        Create a DASK cluster with all `Runtimes` detected by the `RuntimeManager`. Each `Runtime` will host one DASK
        worker and the DASK master (i.e. scheduler) will be started on localhost as implemented in
        `LocalMasterLauncher`.
        >>> from lazycluster import RuntimeManager
        >>> cluster = DaskCluster(RuntimeManager().create_group())
        >>> cluster.start()

        Use different strategies for launching the master and the worker instance by providing custom implementation of
        `MasterLauncher` and `WorkerLauncher`.
        >>> cluster = DaskCluster(RuntimeManager().create_group(),
        ...                       MyMasterLauncherImpl(),
        ...                       MyWorkerLauncherImpl)
        >>> cluster.start()
    """

    DEFAULT_MASTER_PORT = 8786
    PIP_INSTALL_COMMAND = 'pip install -q "dask[complete]"'

    def __init__(self, runtime_group: RuntimeGroup, ports: Optional[List[int]] = None,
                 master_launcher: Optional[MasterLauncher] = None,
                 worker_launcher: Optional[WorkerLauncher] = None):
        """Initialization method.

        Args:
            runtime_group (RuntimeGroup): The `RuntimeGroup` contains all `Runtimes` which can be used for starting the
                                          DASK entities.
            ports (Optional[List[int]]: The list of ports which will be used to instantiate a cluster. Defaults to
                                        list(range(self.DEFAULT_PORT_RANGE_START,
                                                   self.DEFAULT_PORT_RANGE_END)).
            master_launcher (Optional[MasterLauncher]): Optionally, an instance implementing the `MasterLauncher`
                                                        interface can be given, which implements the strategy for
                                                        launching the master instances in the cluster. If None, then
                                                        `LocalMasterLauncher` is used.
            worker_launcher (Optional[WorkerLauncher]): Optionally, an instance implementing the `WorkerLauncher`
                                                        interface can be given, which implements the strategy for
                                                        launching the worker instances. If None, then
                                                        `RoundRobinLauncher` is used.
        """
        super().__init__(runtime_group, ports)

        self._master_launcher = master_launcher if master_launcher else LocalMasterLauncher(runtime_group)
        self._worker_launcher = worker_launcher if worker_launcher else RoundRobinLauncher(runtime_group)

    def get_client(self, timeout: int = 2) -> Client:
        """Get a connected Dask client. 
        
        Args:
            timeout (int): The timeout (s) value passed on to the Dask `Client` constructor. Defaults to 2.

        Raises:
            TimeoutError: If client connection `timeout` expires.
        """
        return Client('localhost:' + str(self.master_port), timeout=timeout)  # Raises TimeoutError



"""Exception module for cluster classes.

Note:
    It is intended that the lazycluster.exception.LazyclusterError should be the parent class of all defined exception
    classes here.
"""

from lazycluster.exceptions import LazyclusterError


class MasterStartError(LazyclusterError):
    """Error indicating that the cluster master instance could not be started successfully. """

    def __init__(self, host: str, port: int):
        """Initialization method.

        Args:
            host (str): The host where the cluster master instance should be started.
            port (int): The port of the cluster master instance.
        """
        self.port = port
        super().__init__('The cluster master instance could not be started successfully on host `' + host + '` on port '
                         + str(port))



"""Runtime management module. This module contains convenient classes for working with `Runtimes` and `RuntimeTasks`."""

from typing import Optional, List, Dict, Generator, Union
from storm import Storm
from multiprocessing import Process
import warnings
import logging
import atexit

from lazycluster import RuntimeTask, Runtime, localhost_has_free_port
from lazycluster import PortInUseError, InvalidRuntimeError, NoPortsLeftError, NoRuntimesDetectedError
import lazycluster._utils as _utils


class RuntimeGroup(object):
    """A group of logically related `Runtimes`.

    The `RuntimeGroup` provides a simplified interface for executing tasks in multiple `Runtimes` or exposing ports
    within a `RuntimeGroup`.

    Examples:
            Execute a `RuntimeTask in a `RuntimGroup`
            >>> # Create instances
            >>> group = RuntimeGroup([Runtime('host-1'), Runtime('host-2')])
            >>> # group = RuntimeGroup(hosts=['host-1', 'host-2'])
            >>> my_task = RuntimeTask('group-demo').run_command('echo Hello Group!')

            >>> # Execute a RuntimeTask in a single Runtime
            >>> single_task = group.execute_task(my_task)
            >>> print(single_task.execution_log[0])

            >>> # Execute a RuntimeTask in the whole RuntimGroup
            >>> task_list = group.execute_task(my_task, broadcast=True)

            >>> # Execute RuntimeTask via RuntimGroup either on a single Runtime
            >>> my_task = RuntimeTask('group-demo').run_command('echo Hello Group!')
            >>> task = group.execute_task(my_task)

            A DB is running on localhost on port `local_port` and the DB is only accessible
            from localhost. But you also want to access the service on the other `Runtimes` on port
            `runtime_port`. Then you can use this method to expose the service which is running on the
            local machine to the remote machines.
            >>> # Expose a port to all Runtimes contained in the Runtime. If a port list is given the next free port is
            >>> # chosen and returned.
            >>> group_port = group.expose_port_to_runtimes(local_port=60000, runtime_port=list(range(60000,60010)))
            >>> print('Local port 60000 is now exposed to port ' + str(group_port) + ' in the RuntimeGroup!')

            A DB is running on a remote host on port `runtime_port` and the DB is only accessible from the remote
            machine itself. But you also want to access the service to other `Runtimes` in the group. Then you can use
            this method to expose the service which is running on one `Runtime` to the whole group.
            >>> # Expose a port from a Runtime to all other ones in the RuntimeGroup. If a port list is given the next
            >>> # free port is chosen and returned.
            >>> group_port = group.expose_port_from_runtime_to_group(host='host-1', runtime_port=60000,
            ...                                                      group_port=list(range(60000,60010)))
            >>> print('Port 60000 of `host-1` is now exposed to port ' + str(group_port) + ' in the RuntimeGroup!')
    """

    _INTERNAL_PORT_MIN = 5800
    _INTERNAL_PORT_MAX = 5999
    _internal_port_range = range(_INTERNAL_PORT_MIN, _INTERNAL_PORT_MAX)

    def __init__(self, runtimes: Optional[List[Runtime]] = None, hosts: Optional[List[str]] = None):
        """ Initialization method.
        
        Args:
            runtimes (Optional[List[Runtime]]): List of `Runtimes`. If not given, then `hosts` must be supplied.
            hosts (Optional[List[str]]): List of hosts, which will be used to instantiate `Runtime` objects. If not
                                         given, then `runtimes` must be supplied.
        Raises:
            ValueError: Either `runtimes` or `hosts` must be supplied. Not both or none.
            InvalidRuntimeError: If a runtime cannot be instantiated via host.
        """
        if not runtimes and not hosts or runtimes and hosts:
            raise ValueError("Either `runtimes` or `hosts` must be supplied. Not both or none.")

        if runtimes:
            self._runtimes = {runtime.host: runtime for runtime in runtimes}
        elif hosts:
            self._runtimes = {host: Runtime(host) for host in hosts}  # Throws InvalidRuntimeError
        self._proc_keys = []
        self._tasks = []
        # Cleanup will be done atexit since usage of destructor may lead to exceptions
        atexit.register(self.cleanup)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        hosts_str = ''
        for runtime in self.runtimes:
            if not hosts_str:
                hosts_str += str(runtime)
            else:
                hosts_str += ', ' + str(runtime)
        return type(self).__name__ + ': ' + hosts_str

    @property
    def hosts(self) -> List[str]:
        """Contained hosts in the group.
        
        Returns:
            List[str]: List with hosts of all `Runtimes`.
        """
        return list(self._runtimes)

    @property
    def runtimes(self) -> List[Runtime]:
        """Contained Runtimes in the group.

        Returns:
            List[Runtime]: List with all `Runtimes`.
        """
        return list(self._runtimes.values())

    @property
    def runtime_count(self) -> int:
        """Get the count of runtimes contained in the group. """
        return len(self.hosts)

    @property
    def task_processes(self) -> List[Process]:
        """Processes from all contained `Runtimes` which were started to execute a `RuntimeTask`.

        Returns:
             List[Process]: Process list.
        """
        processes = []
        for runtime in self._runtimes.values():
            processes.extend(runtime.task_processes)
        return processes

    def print_hosts(self):
        """Print the hosts of the group. """
        print('\n')
        if not self.hosts:
            print('The group is empty!')
            return
        else:
            print('Runtimes contained in Group:')
        for hostname in self.hosts:
            print(self.get_runtime(hostname).class_name + ': ' + hostname)

    def add_runtime(self, host: Optional[str] = None, runtime: Optional[Runtime] = None):
        """Add a `Runtime` to the group either by host or as a `Runtime` object.
        
        Args:
            host (Optional[str]): The host of the runtime. Defaults to None.
            runtime (Optional[Runtime]): The `Runtime` object to be added to the group. Defaults to None.

        Raises:
            ValueError: If the same host is already contained. Or if both host and runtime is given. We refuse
                        the temptation to guess. Also if no argument is supplied.
        """
        if host and runtime:
            raise ValueError('Only host or runtime must be supplied. We refuse the temptation to guess.')
        elif not host and not runtime:
            raise ValueError('Hostname or runtime must be supplied.')
        elif runtime and self.contains_runtime(runtime.host):
            raise ValueError('Runtime with the same host is already contained in the group.')
        elif host and self.contains_runtime(host):
            raise ValueError('Runtime with the same host is already contained in the group.')
        elif host:
            rt = Runtime(host)
        else:
            rt = runtime
        self._runtimes[rt.host] = rt
        print(rt.host + ' added as ' + rt.class_name + ' to the group.')

    def remove_runtime(self, host: str):
        """Remove a runtime from the group by host.
        
        Args:
            host (str): The host of the `Runtime` to be removed from the group.
        """
        if not self.contains_runtime(host):
            warnings.warn('Runtime ' + host + ' is  not contained in the group!')
            return
        del self._runtimes[host]

    def expose_port_to_runtimes(self, local_port: int, runtime_port: Union[int, List[int], None] = None,
                                exclude_hosts: Optional[List[str]] = None) -> int:
        """ Expose a port from localhost to all Runtimes beside the excluded ones so that all traffic on the
        `runtime_port` is forwarded to the `local_port` on the local machine. This corresponds to remote
        port forwarding in ssh tunneling terms. Additionally, all relevant runtimes will be checked if the port is
        actually free.
        
        Args:
            local_port (int): The port on the local machine.
            runtime_port (Union[int, List[int], None]): The port on the runtimes where the `local_port` shall be
                                                        exposed to. May raise PortInUseError if a single port is given.
                                                        If a list is used to automatically find a free port then a
                                                        NoPortsLeftError may be raised. Defaults to `local_port`.
            exclude_hosts (Optional[List[str]]): List with hosts where the port should not be exposed to. Defaults to
                                                 None. Consequently, all `Runtimes` will be considered.

        Returns:
            int: The port which was actually exposed to the `Runtimes`.

        Raises:
            PortInUseError: If `runtime_port` is already in use on at least one Runtime.
            ValueError: Only hosts or `exclude_hosts` must be provided or host is not contained in the group.
        """

        # 1. Determine a free runtime port
        if not runtime_port or not isinstance(runtime_port, list):
            if not runtime_port:
                runtime_port = local_port
            if not self.has_free_port(runtime_port):
                raise PortInUseError(runtime_port, self)

        elif isinstance(runtime_port, list):
            runtime_port = self.get_free_port(runtime_port)  # Raises NoPortsLeftError

        if not exclude_hosts:
            exclude_hosts = []

        if not self.has_free_port(runtime_port, exclude_hosts=exclude_hosts + [Runtime.LOCALHOST]):
            logging.debug('T')
            raise PortInUseError(runtime_port, self)

        for runtime in self.get_runtimes(exclude_hosts=exclude_hosts).values():  # Raises ValueError
            process_key = runtime.expose_port_to_runtime(local_port, runtime_port)
            if process_key:
                self._proc_keys.append(process_key)

        return runtime_port

    def expose_port_from_runtime_to_group(self, host: str, runtime_port: int,
                                          group_port: Union[int, List[int], None] = None) -> int:
        """Expose a port from a `Runtime` to all other `Runtimes` in the `RuntimeGroup` so that all traffic to the
        `group_port` is forwarded to the `runtime_port` of the runtime.
        
        Args:  
            host (str): The host of the `Runtime`.
            runtime_port (int): The port on the runtime.
            group_port (Union[int, List[int], None]): The port on the other runtimes where the `runtime_port` shall be
                                                      exposed to. May raise PortInUseError if a single port is given.
                                                      If a list is used to automatically find a free port then a
                                                      NoPortsLeftError may be raised. Defaults to runtime_port.

        Returns:
            int: The `group_port` that was eventually used.

        Raises:
            ValueError: If host is not contained.
            PortInUseError: If `group_port` is occupied on the local machine.
            NoPortsLeftError: If `group_ports` was given and none of the ports was free.
        """
        if not self.contains_runtime(host):
            raise ValueError('Runtime ' + host + ' is not contained in the group.')

        # 1. Determine a free group port
        if not group_port or not isinstance(group_port, list):
            if not group_port:
                group_port = runtime_port
            if not self.has_free_port(group_port, exclude_hosts=[host]):
                raise PortInUseError(group_port, self)

        elif isinstance(group_port, list):
            group_port = self.get_free_port(group_port)

        # 2. Determine a free port on localhost, since all traffic is tunneled over the local machine
        local_port = None
        if localhost_has_free_port(group_port):
            local_port = group_port
        else:
            # Get the first free port from the local port range
            for current_port in self._internal_port_range:
                if not localhost_has_free_port(current_port):
                    continue
                else:
                    local_port = current_port
                    self._internal_port_range = range(current_port + 1, self._INTERNAL_PORT_MAX)

        if not local_port:
            # self._internal_port_range = None <- does this makes sense? not sure!
            raise NoPortsLeftError()

        # 3. Finally expose the ports
        for runtime in self._runtimes.values():
            if runtime.host == host:
                process_key = runtime.expose_port_from_runtime(runtime_port, local_port)
            else:
                process_key = runtime.expose_port_to_runtime(local_port, group_port)
            if process_key:
                self._proc_keys.append(process_key)

        return group_port

    def execute_task(self, task: RuntimeTask, host: Optional[str] = None,
                     broadcast: bool = False, execute_async: bool = True) -> RuntimeTask or List[RuntimeTask]:
        """Execute a `RuntimeTask` in the whole group or in a single `Runtime`. 
        
        Args:
            task (RuntimeTask): The task to be executed.
            host (str): If task should be executed in ine Runtime. Optionally, the host could be set in order to
                        ensure the execution in a specific Runtime. Defaults to None. Consequently, the least busy
                        `Runtime` will be chosen.
            broadcast (bool): True, if the task will be executed on all `Runtimes`. Defaults to False.
            execute_async (bool): True, if execution will take place async. Defaults to True.

        Returns:
            RuntimeTask or List[RuntimeTask]: Either a single `RuntimeTask` object in case the execution took place
                                              in a single `Runtime` or a list of `RuntimeTasks` if executed in all.

        Raises:
            ValueError: If `host` is given and not contained as `Runtime` in the group.
        """
        if broadcast:
            tasks = []
            for runtime in self.get_runtimes().values():  # Raises ValueError
                print('Start executing task ' + task.name + ' in ' + runtime.class_name + ' ' + runtime.host)
                runtime.execute_task(task, execute_async)
                tasks.append(task)
                self._tasks.append(task)
            return tasks
        else:
            if host:
                if host not in self._runtimes:
                    raise ValueError('The host ' + host + ' is not a valid runtime.')
                self.get_runtime(host).execute_task(task, execute_async)
            else:
                self._get_least_busy_runtime().execute_task(task, execute_async)
            self._tasks.append(task)
            return task

    def join(self):
        """Blocks until `RuntimeTasks` which were started via the `runtime.execute_task()` method terminated. """
        for task in self._tasks:
            task.join()

    def print_log(self):
        """Print the execution logs of each `RuntimeTask` that were executed in the group. """
        for task in self._tasks:
            task.print_log()

    def get_free_port(self, ports: List[int], enforce_check_on_localhost: bool = False) -> int:
        """Return the first port from the list which is currently not in use in the whole group.

        Args:
             ports (List[int]): The list of ports that will be used to find a free port in the group.
             enforce_check_on_localhost (bool): If true the port check will be executed on localhost as well, although
                                                localhost might not be a `Runtime` instance contained in the
                                                `RuntimeGroup`.

        Returns:
            int: The first port from the list which is not yet used within the whole group.

        Raises:
            NoPortsLeftError: If the port list is empty and no free port was found yet.
        """
        localhost_not_in_group = True if Runtime.LOCALHOST not in self._runtimes else False

        for port in ports:

            if enforce_check_on_localhost and localhost_not_in_group and not localhost_has_free_port(port):
                continue

            if not self.has_free_port(port):
                continue

            return port

        raise NoPortsLeftError()

    def has_free_port(self, port: int, exclude_hosts: Union[List[str], str, None] = None) -> bool:
        """Check if a given port is free on `Runtimes` contained in the group. The check can be restricted to a
        specific subset of contained `Runtimes` by excluding some hosts.

        Args:
            port (int): The port to be checked in the group.
            exclude_hosts: (Union[List[str], str, None]): If supplied, the check will be omitted in these `Runtimes`.
                                                          Defaults to None, i.e. not restricted.

        Returns:
            bool: True if port is free on all `Runtimes`, else False.

        Raises:
            ValueError: Only hosts or excl_hostnames must be provided or Hostname is
                        not contained in the group.                     
        """
        is_free = True

        for runtime in self.get_runtimes(exclude_hosts=exclude_hosts).values():  # Raises ValueError
            if not runtime.has_free_port(port):
                print('The port ' + str(port) + ' is currently in use on ' + runtime.class_name + ' ' + runtime.host)
                is_free = False

        return is_free

    def contains_runtime(self, host: str) -> bool:
        """Check if the group contains a `Runtime` identified by host.

        Args:
            host (str): The `Runtime` to be looked for.

        Returns:
            bool: True if runtime is contained in the group, else False.
        """
        return True if host in self.hosts else False

    @property
    def function_returns(self) -> Generator[object, None, None]:
        """Blocks thread until a `RuntimeTasks` finished executing and gives back the return data of the remotely
        executed python functions. The data is returned in the same order as the Tasks were started

        Returns:
            Generator[object, None, None]: The unpickled return data.
        """
        for task in self._tasks:
            for return_data in task.function_returns:
                yield return_data

    def clear_tasks(self):
        """Clears all internal state related to `RuntimeTasks`. """
        # 1: Clean up all state on group level
        self._tasks = []
        self._proc_keys = [proc_key for proc_key in self._proc_keys if not Runtime.is_runtime_task_process(proc_key)]
        # 2: Clean up state in contained `Runtimes`
        for runtime in self._runtimes.values():
            runtime.clear_tasks()

    def get_runtime(self, host: Optional[str] = None) -> Runtime:
        """Returns a runtime based on the host.

        Args:
            host (str): The host which identifies the runtime.

        Returns:
            Runtime: Runtime objects identified by `host`. Defaults to the least busy one, i.e. the one with the fewest
            alive processes that execute a `RuntimeTask`.

        Raises:
            ValueError: Hostname is not contained in the group.
        """
        if host and host not in self.hosts:
            raise ValueError('Hostname not contained in the RuntimeGroup')
        elif host:
            return self._runtimes[host]
        else:
            self._get_least_busy_runtime()

    def get_runtimes(self, include_hosts: Union[str, List[str], None] = None,
                           exclude_hosts: Union[str, List[str], None] = None) -> Dict[str, Runtime]:
        """Convenient methods for getting relevant `Runtimes` contained in the group.

        Args:
            include_hosts: (Union[str, List[str], None] = None): If supplied, only the specified `Runtimes` will be
                                                                 returned. Defaults to None, i.e. not restricted.
            exclude_hosts: (Union[str, List[str], None] = None): If supplied, all `Runtimes` beside the here specified
                                                                 ones will be returned. Defaults to an empty list, i.e.
                                                                 not restricted.
        Raises:
            ValueError: If include_hosts and exclude_hosts is provided or if a host from `include_host` is not contained
                        in the group.
        """
        runtimes = {}

        if include_hosts and exclude_hosts:
            raise ValueError('Only include_hosts or exclude_hosts must be provided. Not both.')

        elif not include_hosts and not exclude_hosts:
            # => Return all runtimes
            return self._runtimes

        elif include_hosts:
            # => Only the ones specified in hosts
            include_hosts = _utils.create_list_from_parameter_value(include_hosts)  # Ensure it's a list
            for hostname in include_hosts:
                if hostname not in self._runtimes:
                    raise ValueError(hostname + ' is not contained in the RuntimeGroup.')
                runtimes.update({hostname: self._runtimes[hostname]})

        elif exclude_hosts:
            # => All but without the ones specified in excluded_hosts
            exclude_hosts = _utils.create_list_from_parameter_value(exclude_hosts)  # Ensure it's a list
            for hostname in self.hosts:
                if hostname not in exclude_hosts:
                    runtimes.update({hostname: self._runtimes[hostname]})

        return runtimes

    def cleanup(self):
        """Release all acquired resources and terminate all processes by calling the
        cleanup method on all contained `Runtimes`. """
        for runtime in self.get_runtimes().values():
            runtime.cleanup()

    def _get_least_busy_runtime(self) -> Runtime:
        """Get the least busiest runtime in terms of alive `RuntimeTask` processes. """
        final_rt = None

        for runtime in self.get_runtimes().values():
            if not final_rt:
                final_rt = runtime
            elif runtime.alive_task_process_count < final_rt.alive_task_process_count:
                final_rt = runtime
        return final_rt


class RuntimeManager(object):
    """The `RuntimeManager` can be used for a simplified resource management, since it aims to automatically detect
    valid `Runtimes` based on the ssh configuration. It can be used to create a `RuntimeGroup` based on the
    automatically detected instances and possibly based on further filters such as GPU availability.
    """

    def __init__(self):
        """Initialization method.

        Raises:
            NoRuntimeDetectedError: If no `Runtime` could be automatically detected.
        """
        runtimes = {}

        # Iterate over ssh configuration entries and look for valid RemoteRuntimes
        ssh_util = Storm()
        for ssh_entry in ssh_util.list_entries(only_servers=True):
            if ssh_entry['host'] in runtimes:
                continue
            if (ssh_entry['host'] == 'localhost' and '127.0.0.1' in runtimes) or (ssh_entry['host'] == '127.0.0.1'
                                                                                  and 'localhost' in runtimes):
                continue
            try:
                runtime = Runtime(ssh_entry['host'])
            except InvalidRuntimeError:
                logging.debug('Host ' + ssh_entry['host'] + ' is not a valid RemoteRuntime.')
                continue
            runtimes.update({runtime.host: runtime})
            print(runtime.host + ' detected as valid RemoteRuntime.')

        try:
            self._group = RuntimeGroup(list(runtimes.values()))
        except ValueError as err:
            raise NoRuntimesDetectedError(err)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        runtimes_str = ''
        for runtime in self._group.runtimes:
            runtimes_str += ', ' + str(runtime)
        return type(self).__name__ + ' detected: ' + runtimes_str

    def create_group(self,
                     include_hosts: Union[str, List[str], None] = None,
                     exclude_hosts: Union[str, List[str], None] = None,
                     gpu_required: bool = False,
                     min_memory: Optional[int] = None,
                     min_cpu_cores: Optional[int] = None,
                     installed_executables: Union[str, List[str], None] = None,
                     filter_commands: Union[str, List[str], None] = None,
                     root_dir: Optional[str] = None) -> RuntimeGroup:
        """Create a runtime group with either all detected `Runtimes` or with a subset thereof.
        
        Args:
            include_hosts (Union[str, List[str], None]): Only these hosts will be included in the `RuntimeGroup`.
                                                         Defaults to None, i.e. not restricted.
            exclude_hosts: (Optional[List[str]] = None): If supplied, all detected `Runtimes` beside the here specified
                                                         ones will be included in the group. Defaults to None, i.e. not
                                                         restricted.
            gpu_required (bool): True, if gpu availability is required. Defaults to False.
            min_memory (Optional[int]): The minimal amount of memory in MB. Defaults to None, i.e. not restricted.
            min_cpu_cores (Optional[int]): The minimum number of cpu cores required. Defaults to None, i.e. not
                                           restricted.
            installed_executables (Union[str, List[str], None]): Possibility to only include `Runtimes` that have an
                                                                 specific executables installed. See examples.
            filter_commands (Union[str, List[str], None]): Shell commands that can be used for generic filtering.
            root_dir (Optional[str]): The directory which shall act as root one. Defaults to None.
                                      Consequently, a temporary directory will be created and used as root directory. If
                                      the root directory is a temporary one it will be cleaned up either `atexit` or
                                      when calling `cleanup()` manually.

        Note:
            The filter criteria are passed through the `check_filter()` method of the `Runtime` class. See its
            documentation for further details and examples.

        Returns:
            RuntimeGroup: The created `RuntimeGroup`.
        
        Raises:
            ValueError: Only hosts or excluded_hosts must be provided or Hostname is not contained in the group.
            NoRuntimesError: If no `Runtime` matches the filter criteria.
        """
        runtimes_dict = self._group.get_runtimes(include_hosts, exclude_hosts)  # Raises ValueError

        final_runtimes = []

        for runtime in runtimes_dict.values():
            if runtime.check_filter(gpu_required, min_memory, min_cpu_cores, installed_executables, filter_commands):
                runtime._root_dir = root_dir
                final_runtimes.append(runtime)

        try:
            group = RuntimeGroup(final_runtimes)
        except ValueError as e:
            raise NoRuntimesDetectedError(e)

        return group



"""Exception module."""

from typing import Optional


class LazyclusterError(Exception):
    """Basic exception class for `lazycluster` library errors. """

    def __init__(self, msg: str, predecessor_excp: Optional[Exception] = None):
        """Constructor method.

         Args:
             msg (str): The error message.
         """
        self.msg = msg
        self.predecessor_excp = predecessor_excp

        super().__init__(str(self))

    def __str__(self):
        if self.predecessor_excp:
            return self.msg + ' Predecessor Exception: ' + str(self.predecessor_excp)
        return self.msg


class InvalidRuntimeError(LazyclusterError):
    """Error indicating that a `Runtime` can not be instantiated properly. """

    def __init__(self, host: str):
        """Constructor method.

        Args:
            host (str): The host which cannot be instantiated as `Runtime`.
        """
        self.host = host

        super().__init__('No runtime could be instantiated for host :' + self.host)


class NoRuntimesDetectedError(LazyclusterError):
    """Error indicating that no `Runtime` could be detcted automatically by a `RuntimeManager` for example. """
    def __init__(self, predecessor_excp: Optional[Exception] = None):
        super().__init__('No Runtimes automatically detected.', predecessor_excp)


class PortInUseError(LazyclusterError):
    """Error indicating that a port is already in use in a `RuntimeGroup` or on the local machine. """

    def __init__(self, port: int, group: Optional['RuntimeGroup'] = None, runtime: Optional['Runtime'] = None):
        """Constructor method.

        Args:
            port (int): The port in use.
            group (Optional[RuntimeGroup]): The group object where the port is in use.
            runtime (Optional[Runtime]: The runtime object where the port is in use.
        """
        self.port = port
        self.group = group
        self.runtime = runtime

        if group:
            msg = 'Port ' + str(port) + ' is already in use in at least one `Runtime` in the group or on the local ' \
                  'machine. Group: ' + str(group)
        elif runtime:
            msg = 'Port ' + str(port) + ' is already in use in ' + runtime.type_str + ' ' + runtime.host + \
                  '. Runtime: ' + str(runtime)
        else:
            msg = 'Port ' + str(port) + ' is already in use.'

        super().__init__(msg)


class NoPortsLeftError(LazyclusterError):
    """Error indicating that there are no more ports left from the given port list. """

    def __init__(self):
        """Constructor method."""
        msg = 'No free port could be determined. No more ports left in port list.'
        super().__init__(msg)



from typing import List, Union


def get_remaining_ports(ports: List[int], last_used_port: int) -> List[int]:
    """Get a new list with the remaining ports after cutting out all ports until and incl. the last used port.

    Args:
        ports (List[int]): The port list to be updated.
        last_used_port (int): The last port that was actually used. All ports up this one and including this one will
                              be removed.
    Returns:
        List with remaining ports
    """
    skip = True
    final_port_list = []
    for port in ports:
        if skip:
            if port == last_used_port:
                skip = False
            continue
        final_port_list.append(port)
    return final_port_list


def create_list_from_parameter_value(value: Union[object, List[object], None]) -> list:
    """Many methods can either take a single value, a list or None as input. Usually we want to iterate all given
    values. Therefore, this method ensures that a list is always returned.

    Args:
        value (Union[object, List[object], None])): The value that will be mapped onto a list representation.

    Returns:
        List[object]: Either an empty list or a list with all given objects.

    Examples:
        value = None    => []
        value = []      => []
        value = 5       => [5]
        value = [5,6]   => [5,6]
    """
    if not value:
        # => create empty list
        return []
    if not isinstance(value, list):
        # => create list with single value
        return [value]
    # => value is already a list
    return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created based on: https://github.com/kennethreitz/setup.py/blob/master/setup.py
# Alternative: https://github.com/seanfisk/python-project-template/blob/master/setup.py.tpl
import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'lazycluster'
MAIN_PACKAGE = 'lazycluster'  # Change if main package != NAME
DESCRIPTION = 'Distributed machine learning made simple.'
URL = 'https://github.com/ml-tooling/lazycluster.git'
EMAIL = 'jan.kalkan@mailbox.org'
AUTHOR = 'Jan Kalkan'
REQUIRES_PYTHON = '>=3.6'
VERSION = None  # Only set version if you like to overwrite the version in __version__.py

# Please define the requirements within the requirements.txt

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

# Check if version is right
if (sys.version_info[:1] == 3 and sys.version_info[:2] < (3, 5)):  # sys.version_info[:2] < (2, 7) or
    raise Exception('This package needs Python 3.6 or later.')  # 2.7,

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!

#

try:
    # convert markdown to rst format
    import pypandoc

    long_description = pypandoc.convert(here, 'README.md', 'rst')
    long_description = long_description.replace("\r", "") 
except:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()

# Read the requirements.txt and use it as the setup.py requirements
with io.open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = [line.rstrip() for line in f.readlines()]

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, MAIN_PACKAGE, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class DeployCommand(Command):
    """Support setup.py upload."""

    # Note: To use the 'deploy' functionality of this file, you must:
    #   $ pip install twine

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description="%s\n\nRequirements:\n%s" % (long_description, requirements),
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    license='Apache License 2.0',
    packages=find_packages(exclude=('tests',)),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        # TODO: update this list to match your application: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        # Not python 2 compatible right now
        # 'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    # $ setup.py publish support.
    cmdclass={
        'deploy': DeployCommand
    },
)
